#include <KeyMatrix.h>
#include <KeyMatrixPhone.h>
#include <LiquidCrystal.h>

#include "pitches.h"

//User configurable Elements:
int giveClues = 1;    // If 1 clues will be given (every 5 minutes stuck in a round), if 0 no clues will be given
int wrongAnswerPenalty = 0; //Number of minutes to penalise teams for wrong answer (0 = no penalty, 1 = 1 minute penalty) 
static volatile int countDownClock = 1200;  //Time in seconds (3300 = 55 minutes, 2700 = 45 minutes)

//Keys go here: Can be 1-8 digits long. Must be 3 keys
String Key1 = "16214364";
String Key2 = "43241";
String Key3 = "8425";

//String Key1 = "5";
//String Key2 = "9801";
//String Key3 = "23466432";

//Non-user configurable elements below

int timeInRound = countDownClock;
int LCD_E = 4;
int LCD_RS = 5;
int LCD_DB4 = 14;
int LCD_DB5 = 15;
int LCD_DB6 = 16;
int LCD_DB7 = 17;

int keypad1 = 12; 
int keypad2 = 11;
int keypad3 = 10;
int keypad4 = 9;
int keypad5 = 8;
int keypad6 = 7;
int keypad7 = 6;

int stateCounter = 1;
int currDigit = 1;    //The digit we are trying to enter
int enteringCodeFlag = 0; //If we have started entering the code set flag to 1 (prevents X's from overwriting digits)
String codeString;

int melody_fail[] = {
  NOTE_G3, NOTE_C3
};

int melody_succeed[] = {
  NOTE_C3, NOTE_C4, NOTE_C5, NOTE_C6
};

// note durations: 4 = quarter note, 8 = eighth note, etc.:
int noteDurationsFail[] = {
  2, 1
};

int noteDurationsSucceed[] = {
  8, 8, 8, 8
};

AK_304(keypad, keypad1, keypad2, keypad3, keypad4, keypad5, keypad6, keypad7);

int readKeypad(void);
void playSFXFail(void);
void playSFXSucceed(void);
char checkCode(String codeString);
int displayRoundTime(void);
void setupCountDownTimer(void);
String getCurrKey(void);
  

LiquidCrystal lcd(LCD_RS, LCD_E, LCD_DB4, LCD_DB5, LCD_DB6, LCD_DB7);

void setup() {
  //playSFXSucceed();
  //playSFXFail();
  Serial.begin(9600);
  lcd.begin(8, 2);
  
  lcd.setCursor(1, 0);
  lcd.print("Escape");
  lcd.setCursor(2, 1);
  lcd.print("Tute");
  delay(2000);  //2 Second intro then start countdown
  setupCountDownTimer();

}

void loop() {
    char keyPadValue;  
  while(1){
   
    displayRoundTime();
    displayCodeOrClue();

    keyPadValue = readKeypad();
    if(keyPadValue == '*'){ //'*' pressed - clear display
      currDigit = 1;    
      enteringCodeFlag = 0; 
      codeString = "";
    }
    else{
      if(keyPadValue != 'X'){
        //Check which digit we are up to (and has been revealed) and write to that position
        lcd.setCursor(currDigit-1, 1); 
        lcd.print(keyPadValue);
        enteringCodeFlag = 1;     //User has started to enter code
        codeString += keyPadValue;  //Add to the code
        currDigit++;              //Allow user to enter next digit      
      }
    }
    if(getCurrKey().length() == codeString.length()){
      //A full string has been entered, now need to check if it is correct
      if(1 == checkCode(codeString)){
        //Success
        lcd.setCursor(0, 1);
        lcd.print("Correct!");
        playSFXSucceed();
        stateCounter++;
        Serial.print("Moved to stage:");
        Serial.println(stateCounter);
        timeInRound = countDownClock;
      }
      else{
        //Failure
        lcd.setCursor(0, 0);
        lcd.print("Try");
        lcd.setCursor(0, 1);
        lcd.print("Again   ");        
        //Apply penalty for incorrect answers (may be no penalty if wrongAnswerPenalty = 0)
        countDownClock -= (wrongAnswerPenalty * 60);
        if(countDownClock < 0){
          countDownClock = 0;
        }
        playSFXFail();
      }
      codeString = "";           
      enteringCodeFlag = 0;      
    }

    if(0 == countDownClock){
      //Ran out of time - game over!
      lcd.setCursor(2, 0); 
      lcd.print("  GAME");
      lcd.setCursor(0, 1); 
      lcd.print("    OVER");
      playSFXFail();
      playSFXFail();
      playSFXFail();
      while(1){}      
    }

    if(4 == stateCounter){
      //Team has completed all codes
      lcd.setCursor(0, 1); 
      lcd.print("ESCAPED!");
      lcd.setCursor(0, 0); 
      lcd.print("  ");
      while(1){}      
    }

    int timeStuck = timeInRound - countDownClock; //Every 300 seconds reveal another digit, but never reveal the last digit and not if the user has started to enter in a password
    if(0 == enteringCodeFlag){  //Not started entering a code
      int keyLength = getCurrKey().length();
      //if less than 300 currDigit will always be 1, otherwise needs to be a multiple of 300
      currDigit = (timeStuck / 300) + 1;
      if(currDigit > keyLength){
        currDigit = keyLength;
      }
    }
    
    delay(100);
    Serial.println(codeString);   
  }
}

void setupCountDownTimer(void){
  //Uses timer 1 (16MHz) to generate a 1 second interrupt to count down.
  //Clock is 16MHz
  //Use CTC mode (Clear Timer on Compare Match with OCR1A)
  TCCR1A = 0x00;  //00000000 //Dont connect outputs, CTC mode
  TCCR1B = 0x0C; //00001101 Dont use output compare, CTC mode, 256 prescaler (100)
  TCNT1 = 0;
  TIMSK1 = 0x02;  //Enable interrupt on Match
  OCR1A = 62500; //Will give a 1s period
  //OCR1A = 3000; //Will give a ~50ms period
}


ISR(TIMER1_COMPA_vect) {
   if(countDownClock != 0){
      countDownClock--;
   }
}

int displayRoundTime(void){
  int tmpTimeFrac;
  //Display Round information
  lcd.setCursor(0, 0);
  lcd.print("R");
  lcd.print(stateCounter);
  lcd.print(" ");

  //Display Time information  
  tmpTimeFrac = countDownClock/60;
  if(tmpTimeFrac < 10)
    lcd.print("0");
  lcd.print(tmpTimeFrac);
  lcd.print(":");
  
  tmpTimeFrac = countDownClock%60;
  if(tmpTimeFrac < 10)
    lcd.print("0");    
  lcd.print(tmpTimeFrac);
}

String getCurrKey(void){  //Returns the key string corresponding to the current round that you are trying to solve
  if(1 == stateCounter)
    return Key1;
  if(2 == stateCounter)
    return Key2;
  if(3 == stateCounter)
    return Key3;        
}

int displayCodeOrClue(void){
  if(1 == giveClues){
      int timeStuck = timeInRound - countDownClock; //Seconds stuck for - every 300 seconds (5 min) gives a new clue
    //Clean up screen incase correct or try-again message was shown
    int keyLength = getCurrKey().length();
  
    lcd.setCursor(keyLength,1);
    lcd.print("       ");

    if(0 == enteringCodeFlag){ //Player is not entering a code
      //Will display **** as the code, unless the user has been trying for more than 5 minutes, every 5 minutes one additional extra digit will be revealed     
      lcd.setCursor(0, 1);  //Display from this point
      //Based on the value of currDigit print out characters we are willing to reveal (always keeping at least one secret)
     //Update the value of codeString accordingly
     lcd.print(getCurrKey().substring(0,currDigit-1));
     for(int i = currDigit; i <= keyLength; i++){
       lcd.print("*");
     }
      codeString = getCurrKey().substring(0,currDigit - 1);
    }
  }
}

void playSFXFail(void){
  for (int thisNote = 0; thisNote < 2; thisNote++) {

    // to calculate the note duration, take one second divided by the note type.
    //e.g. quarter note = 1000 / 4, eighth note = 1000/8, etc.
    int noteDuration = 1000 / noteDurationsFail[thisNote];
    tone(3, melody_fail[thisNote], noteDuration);

    // to distinguish the notes, set a minimum time between them.
    // the note's duration + 30% seems to work well:
    int pauseBetweenNotes = noteDuration * 1.30;
    delay(pauseBetweenNotes);
    // stop the tone playing:
    noTone(3);
  }
}


void playSFXSucceed(void){  
  for (int thisNote = 0; thisNote < 4; thisNote++) {

    // to calculate the note duration, take one second divided by the note type.
    //e.g. quarter note = 1000 / 4, eighth note = 1000/8, etc.
    int noteDuration = 1000 / noteDurationsSucceed[thisNote];
    tone(3, melody_succeed[thisNote], noteDuration);

    // to distinguish the notes, set a minimum time between them.
    // the note's duration + 30% seems to work well:
    int pauseBetweenNotes = noteDuration * 1.30;
    delay(pauseBetweenNotes);
    // stop the tone playing:
    noTone(3);
  }
}

int readKeypad(void){
  if (keypad.pollEvent()) {
        // event data is stored on keypad.event
        if (keypad.event.type == KM_KEYDOWN){
            Serial.println(keypad.event.c);
            tone(3, NOTE_C4, 50); 
            return keypad.event.c;
        }       
  }
  return 'X';
}

char checkCode(String codeString){
  if(getCurrKey() == codeString){
    return 1;
  }
  else{
    return 0;
  }
}

